#!/usr/bin/env python3
import os
from random import choice
from contextlib import contextmanager

from bsddb3.db import DB
from bsddb3.db import DBEnv
from bsddb3.db import DB_BTREE
from bsddb3.db import DB_CREATE
from bsddb3.db import DB_INIT_LOG
from bsddb3.db import DB_INIT_TXN
from bsddb3.db import DB_INIT_MPOOL
from bsddb3.db import DB_MULTIVERSION
from bsddb3.db import DB_RECOVER_FATAL

from json import dumps
from json import loads


__version__ = '1.99.2'


ID_LENGTH = 128


VERTEX = 'v'
EDGE = 'g'
DATA = 'd'
INCOMINGS = 'i'
OUTGOINGS = 'o'
LABEL = 'l'
START = 's'
END = 'e'
ID = 'x'

ALPHANUM = 'azertyuiopqsdfghjklmwxcvbnAZERTYUIOPQSDFGHJKLLMWXCVBN1234567890'


def random_identifier():
    return ''.join([choice(ALPHANUM) for x in range(ID_LENGTH)])


class Element:

    def id(self):
        return self._id

    def __setitem__(self, key, value):
        self._data[key] = value

    def __getitem__(self, key):
        return self._data[key]

    def get(self, key, default):
        return self.data.get(key, default)

    def save(self):
        k = '{}{}'.format(self.NAME, self._id).encode('ascii')
        v = dumps(self._dump())
        self._txn.db.put(k, v, self._txn._txn)

    def delete(self):
        k = '{}{}'.format(self.NAME, self._id).encode('ascii')
        self._txn.db.delete(k, self._txn._txn)


class Vertex(Element):

    __slots__ = ('_incomings', '_outgoings', '_data', '_id', '_txn')

    NAME = VERTEX

    def __init__(self, txn, id):
        self._incomings = list()
        self._outgoings = list()
        self._data = dict()
        self._id = id
        self._txn = txn

    def _dump(self):
        out = dict()
        out[INCOMINGS] = self._incomings
        out[OUTGOINGS] = self._outgoings
        out[DATA] = self._data
        out[ID] = self._id
        return out

    @classmethod
    def _load(cls, graph, values):
        node = cls(graph, values[ID])
        node._incomings = values[INCOMINGS]
        node._outgoings = values[OUTGOINGS]
        node._data = values[DATA]
        return node

    def incomings(self, label=None):
        if not label:
            edges = self._incomings
        else:
            edges = filter(lambda x: x[1] == label, self._incomings)
        return map(lambda x: self._txn.edge(x[0]), edges)

    def outgoings(self, label=None):
        if not label:
            edges = self._outgoings
        else:
            edges = filter(lambda x: x[1] == label, self._outgoings)
        return map(lambda x: self._txn.edge(x[0]), edges)


class Edge(Element):

    __slots__ = ('_id', '_start', '_label', '_end', '_data', '_txn')

    NAME = EDGE

    def __init__(self, txn, id, start, label, end):
        self._id = id
        self._start = start
        self._label = label
        self._end = end
        self._data = dict()
        self._txn = txn

    def _dump(self):
        out = dict()
        out[ID] = self._id
        out[START] = self._start
        out[LABEL] = self._label
        out[END] = self._end
        out[DATA] = self._data
        return out

    @classmethod
    def _load(cls, graph, values):
        edge = cls(graph, values[ID], values[START], values[LABEL], values[END])
        edge._data = values[DATA]
        return edge

    def start(self):
        return self._txn.vertex(self._start)

    def end(self):
        return self._txn.vertex(self._end)

    def label(self):
        return self._label

    def delete(self):
        start = self.start()
        start._outgoings.remove([self._id, self._label])
        start.save()
        end  = self.end()
        end._incomings.remove([self._id, self._label])
        end.save()
        super().delete()


class Transaction:

    __slots__ = ('_txn', '_graph', 'db')

    def __init__(self, graph):
        self._txn = graph.env.txn_begin()
        self._graph = graph
        self.db = graph.db

    def create_vertex(self):
        while True:
            id = random_identifier()
            k = '{}{}'.format(VERTEX, id).encode('ascii')
            if self.db.get(k, txn=self._txn):
                continue
            else:
                break
        vertex = Vertex(self, id)
        vertex.save()
        return vertex

    def vertex(self, id):
        k = '{}{}'.format(VERTEX, id).encode('ascii')
        v = self.db.get(k, txn=self._txn)
        if v:
            return Vertex._load(self, loads(v, encoding='utf-8'))
        return None

    def create_edge(self, start, label, end):
        while True:
            id = random_identifier()
            k = '{}{}'.format(VERTEX, id).encode('ascii')
            if self.db.get(k, txn=self._txn):
                continue
            else:
                break
        edge = Edge(self, id, start.id(), label, end.id())
        # edges are stored with label to improve edge filtering
        end._incomings.append((edge.id(), edge.label()))
        end.save()
        start._outgoings.append((edge.id(), edge.label()))
        start.save()
        edge.save()
        return edge

    def edge(self, id):
        k = '{}{}'.format(EDGE, id).encode('ascii')
        v = self.db.get(k, txn=self._txn)
        if v:
            return Edge._load(self, loads(v, encoding='utf-8'))
        return None

    def set(self, key, value):
        self.db.put(key.encode('ascii'), dumps(value), txn=self._txn)

    def get(self, key):
        v = self.db.get(key.encode('ascii'), txn=self._txn)
        if v:
            return loads(v, encoding='utf-8')
        return None
    
    def commit(self):
        self._txn.commit()

    def rollback(self):
        self._txn.abort()


class Graph:

    def __init__(self, path, cache_size=(0, 512), max_cache_size=(0, 512)):
        if not os.path.exists(path):
            os.makedirs(path)
        self.env = DBEnv()
        self.env.set_cache_max(*max_cache_size)
        self.env.open(
            path,
            DB_INIT_LOG | DB_INIT_TXN | DB_RECOVER_FATAL | DB_CREATE | DB_INIT_MPOOL,
            0
        )
        self.env.set_cachesize(*cache_size)

        self.db = DB(self.env)
        txn = self.txn()
        self.db.open('ajgu', 'ajgu', DB_BTREE, DB_CREATE | DB_MULTIVERSION, 0, txn=txn._txn)
        txn.commit()

    def close(self):
        self.db.close()
        self.env.close()

    def txn(self):
        return Transaction(self)

    @contextmanager
    def transaction(self):
        txn = self.txn()
        try:
            yield txn
        except Exception as exc:
            txn.rollback()
            raise exc
        else:
            txn.commit()


if __name__ == '__main__':
    import sys
    import asyncio
    import websockets
    graph = Graph(sys.argv[1])
    print('running on localhost:8765')

    def serve(websocket, uri):
        while websocket.open:
            msg = yield from websocket.recv()
            query, args, kwargs = loads(msg)
            args.insert(0, graph)
            # retrieve function
            compiled = compile(query, '<string>', 'exec')
            context = dict()
            exec(compiled, {}, context)
            name, query = context.popitem()
            try:
                v = query(*args, **kwargs)
            except:
                websocket.send(dumps({'type': 'error'}))
            else:
                websocket.send(dumps({'type': 'return', 'value': v}))

    start_server = websockets.serve(serve, 'localhost', 8765)
    asyncio.get_event_loop().run_until_complete(start_server)
    asyncio.get_event_loop().run_forever()
